use super::{
    req_props::{Props, ReqProps, Response},
    service::Service,
};
use crate::utils::path;
use std::sync::Arc;
use tokio::{sync::Mutex, task::JoinSet};

pub const DEFAULT_DEPTH: u16 = 3;

pub struct Enumerator<'a> {
    service: &'a Service,
    path: String,
    depth: u16,
    properties: Vec<Props>,
}

impl<'a> Enumerator<'a> {
    pub fn new(service: &'a Service) -> Self {
        Enumerator {
            service,
            path: String::new(),
            depth: DEFAULT_DEPTH,
            properties: Vec::new(),
        }
    }

    pub fn set_path(mut self, path: String) -> Self {
        self.path = path;
        self
    }

    pub fn set_depth(mut self, depth: u16) -> Self {
        self.depth = depth;
        self
    }

    pub fn get_properties(mut self, properties: Vec<Props>) -> Self {
        self.properties.extend(properties);
        self
    }

    pub async fn enumerate(&self) -> Result<(Vec<Response>, Vec<Response>), std::io::Error> {
        let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel();
        let files = Arc::new(Mutex::new(Vec::new()));
        let folders = Arc::new(Mutex::new(Vec::new()));
        let service = Arc::from(self.service.clone());
        let tasks_active = Arc::new(Mutex::new(0));

        tx.send(self.path.clone());

        let mut taskSet = JoinSet::new();

        loop {
            if let Ok(path) = rx.try_recv() {
                let current_depth = path::get_depth(&path);
                *tasks_active.lock().await += 1;
                let tx_clone = tx.clone();
                let files_clone = Arc::clone(&files);
                let folders_clone = Arc::clone(&folders);
                let service_clone = Arc::clone(&service);
                let properties = self.properties.clone();
                let depth = self.depth.clone();
                let tasks_active_clone = Arc::clone(&tasks_active);

                let task = taskSet.spawn(async move {
                    let res = ReqProps::new(&service_clone)
                        .set_path(path.clone())
                        .set_depth(depth)
                        .get_properties(properties)
                        .send()
                        .await
                        .unwrap();

                    for obj in res.responses {
                        if obj.is_dir() {
                            // Avoid enumerating the same folder multiple times
                            if obj.abs_path() != &path {
                                // depth deeper than current + self.depth
                                if obj.path_depth() >= current_depth + depth {
                                    tx_clone.send(obj.abs_path().to_owned()).unwrap();
                                }
                                folders_clone.lock().await.push(obj);
                            }
                        } else {
                            files_clone.lock().await.push(obj);
                        }
                    }
                    *tasks_active_clone.lock().await -= 1;
                });
            } else if *tasks_active.lock().await <= 0 {
                break;
            }
        }

        taskSet.join_all().await;

        Ok((
            Arc::try_unwrap(files).unwrap().into_inner(),
            Arc::try_unwrap(folders).unwrap().into_inner(),
        ))
    }
}
